

--------------------------------------------------------------------------------------------------
--										Startup function
--------------------------------------------------------------------------------------------------

function JFCloudOnOpen(posx, posy)
	local cloud = JFCloud:initWithModel(nil)
	cloud:spawnAtPosition(CGPoint(posx, posy))
end

--[[
--------------------------------------------------------------------------------------------------
--										JFCloud
--------------------------------------------------------------------------------------------------
waxClass{"JFCloud", JFActorChaser}

nextIdentifier = 1.0

function initWithModel(self, model)
	self = self.super:initWithModel(model)
	if self then
		self.identifier = nextIdentifier;
		nextIdentifier = nextIdentifier + 1.0

		self:rootNode():setScale(0.0)

		self:setupStateMachine()
		
		self.isMovingToVillager = false
		
		self.numberOfVillagersToAttack = 3
		self.numberOfVillagersAttacked = 0
		
		self.animations = 
		{
			"FX_CLOUD_RAIN",
			"FX_CLOUD_RAINBOW",
			"FX_CLOUD_LIGHTNING"
		}
	end
	return self
end


function getRandomAnimation(self)
	return self.animations[math.random(#self.animations)]
end


function setupStateMachine(self)
	self:stateMachine():addState(JFCloudStateArriving:initWithCloud(self))
	self:stateMachine():addState(JFCloudStateChasing:initWithCloud(self))
	self:stateMachine():addState(JFCloudStateJudging:initWithCloud(self))
	self:stateMachine():addState(JFCloudStateDying:initWithCloud(self))
end


function spawnAtPosition(self, tilePosition)
	-- get the services requried
	local gos = JFServiceLocator:serviceOfType(JFGameObjectService:class()) -- game object service
	local map = JFServiceLocator:serviceOfType(JFMapService:class()) -- map service
	
	-- set position
	self:setIsoPos(tilePosition)

	-- add to the game object service and the map
	gos:addGameObject(self)
	map:addActor(self)
	
	-- put the cloud into the arriving state
	self:stateMachine():setCurrentState_withMessage(JFCloudStateArriving:class(), nil)
end


function classId(self)
    return JFUtility:convertClassToClassId(JFCloud:class())
end


function ident(self)
	return self.identifier
end


function untrackVillager(self)
	self.isMovingToVillager = false
	self.villager = nil
end


function trackVillager(self, villager)
	if not self.villager == nil then
		self:untrackVillager()
	end
	
	self.isMovingToVillager = true
	self.villager = villager
	
end



function handleMessage(self, message)
	self.super:handleMessage(message)
	
	if message:name() == "VILLAGER_INVALIDATED" then
		self.isMovingToVillager = false
	end
end



function update(self, dt)
	self.super:update(dt)
	self:moveToActor(self.villager, dt)
end


function moveToActor(self, villager, dt)
	if villager == nil then
		return
	end
	if self.isMovingToVillager == false then
		self:rootNode():setPosition(villager:rootNode():position())
		self:updateSortingDepth()
	end
	
	-- map service
	local mapService = JFServiceLocator:serviceOfType(JFMapService:class())
	
	-- get the positions of the villager and cloud
	local actorPosition = self.villager:screenPos()
	local myPosition = self:screenPos()
	
	-- get the normalized direction from the difference
	local difference = CGPoint(actorPosition.x - myPosition.x, actorPosition.y - myPosition.y)
	local length = math.sqrt(difference.x*difference.x+difference.y*difference.y)
	local dir = CGPoint(difference.x / length, difference.y / length)
	
	-- get the movement amount
	
	
	local magnitude = villager:walkSpeed() * 1.1 * dt
	local movement = CGPoint(dir.x * magnitude, dir.y * magnitude)
	
	-- get the new position
	local worldPos = self:rootNode():position()
	local newPosition = CGPoint(worldPos.x + movement.x, worldPos.y + movement.y)
	
	-- check the distance to the target and if the cloud is going to overshoot, set to actor position
	if length <= magnitude then
		newPosition = villager:rootNode():position()
		if self.isMovingToVillager == true then
			self:stateMachine():currentState():changeState(JFCloudStateJudging)
		end
		self.isMovingToVillager = false
	end
	
	-- set the new position of the cloud, then sanity check the villager is still valid
	self:rootNode():setPosition(newPosition)
	self:updateSortingDepth()	
end



--------------------------------------------------------------------------------------------------
--										JFCloudState
--------------------------------------------------------------------------------------------------
waxClass{"JFCloudState", JFStateBase}

function initWithCloud(self, acloud)
	self = self.super:init()
	if self then
		self.cloud = acloud
	end
	return self
end


function getCloud(self)
	return self.cloud
end


function changeState(self, stateClass)
	-- convenience function so I don't have to write this out every time
	self.cloud:stateMachine():setCurrentState_withMessage(stateClass:class(), nil)
end


-- takes care of adding the cloud as sender
function createMessage(self, messageName, messageData)
	
	assert(messageName ~= nil)
	assert(self.cloud ~= nil)
	
	if messageData == nil then
		messageData = { }
	end
	
	messageData["sender"] = self.cloud
	
	local message = JFMessage:messageWithName(messageName)
	message:setData(messageData)
	
	return message
	
end


function sendMessageToVillager(self, messageName, messageData)

	local message = self:createMessage(messageName, messageData)
	
	local villager = self:getCloud():getVillager()
	villager:handleMessage(message)
	
end


--------------------------------------------------------------------------------------------------
--										JFCloudStateArriving
--------------------------------------------------------------------------------------------------
waxClass{"JFCloudStateArriving", JFCloudState}

function onEnter_withMessage(self, previousState, message)
	local cloud = self:getCloud()
	
	-- set costume
	cloud:wardrobe():playAnimation("FX_CLOUD_IDLE")
	cloud:updateSortingDepth()
	
	-- create pop in action
	local actionScale = CCScaleTo:actionWithDuration_scale(1.0, 1.0)
	local actionFinish = CCCallFunc:actionWithTarget_selector(self, "animationFinished")
	
	-- build sequences
	local sequence = CCSequence:actionOne_two(actionScale, actionFinish)
	
	-- run sequence
	cloud:rootNode():runAction(sequence)
end


function animationFinished(self)
	self:changeState(JFCloudStateChasing)
end


--------------------------------------------------------------------------------------------------
--										JFCloudStateChasing
--------------------------------------------------------------------------------------------------
waxClass{"JFCloudStateChasing", JFCloudState}

function onFirstRun_withMessage(self, previousState, message)
	self.super:onFirstRun_withMessage(previousState, message)
	self:getCloud():chaseVillager(nil)
end



function handleMessage(self, message)
	if message:name() == "STARTED_CHASING_VILLAGER" then
		local cloud = self:getCloud()
		local villager = cloud:getVillager()
	
		cloud:trackVillager(villager)
		if(villager:stateMachine():stateOfType(JFVillagerStateRunningAway:class()) == nil) then
			local state = JFVillagerStateRunningAway:initWithVillager(villager)
			villager:stateMachine():addState_forClass(state, JFVillagerStateRunningAway:class())
			io.write("added state to villager")
		end
	
		local message = self:createMessage(kMessage_CloudStartedChasing, { })
	
		villager:stateMachine():setCurrentState_withMessage(JFVillagerStateRunningAway:class(), message)	
	end
	
	if message:name() == "VILLAGER_INVALIDATED" then
		self:getCloud():chaseVillager(nil)
		return
	end
	
	if message:name() == "NO_VILLAGERS" then
		self:onNoVillagers()
		return
	end
end


function onNoVillagers(self)
	self:changeState(JFCloudStateDying)
end


--------------------------------------------------------------------------------------------------
--										JFCloudStateJudging
--------------------------------------------------------------------------------------------------
waxClass{"JFCloudStateJudging", JFCloudState}


function onEnter_withMessage(self, previousState, message)

	self.super:onEnter_withMessage(previousState, message)

	local actionAnimateCloud = CCCallFunc:actionWithTarget_selector(self, "playCloudAnimation")
	local actionDelay = CCDelayTime:actionWithDuration(0.1)
	--local actionAnimateVillager = CCCallFunc:actionWithTarget_selector(self, "playVillagerAnimation")
	
	local seqTwo = CCSequence:actionOne_two(actionAnimateCloud, actionDelay)
	--local seqTwo = CCSequence:actionOne_two(seqOne, actionAnimateVillager)

	local cloud = self:getCloud();

	cloud:rootNode():runAction(seqTwo)	
	
	self.wasLocked = self:getCloud():getVillager():stateMachine():currentState():locked()
	self:getCloud():getVillager():stateMachine():currentState():setLocked(true)
	
	self:sendMessageToVillager(kMessage_CloudStartedAttacking, { })
	
end


function onExit_withMessage(self, nextState, message)
	self.super:onExit_withMessage(nextState, message)
	
	local cloud = self:getCloud()
	local villager = cloud:getVillager()
	
	villager:stateMachine():currentState():setLocked(self.wasLocked)
	
	self:sendMessageToVillager(kMessage_CloudFinishedAttacking, { })
	
	-- after this call, cloud.villager is nil so anything needing the villager must be
	-- done before this call
	cloud:finishChasingVillager()
	cloud:untrackVillager()
end


function playCloudAnimation(self)
	local cloud = self:getCloud()
	self.animation = cloud:getRandomAnimation()
	local animation = cloud:wardrobe():playAnimation(self.animation)
	animation:addCallbackOnFinished_withSelector(self, "onFinished:withAnimation:")
	
	
end

function playVillagerAnimation(self)
	local villager = self:getCloud():getVillager()
	villager:playBoostAnimation()
end

function onFinished_withAnimation(self, animationName, animation)
	if animationName == self.animation then
		
		self:getCloud():wardrobe():playAnimation("FX_CLOUD_IDLE")
	
		local cloud = self:getCloud()
		
		cloud.numberOfVillagersAttacked = cloud.numberOfVillagersAttacked + 1
		
		if cloud.numberOfVillagersToAttack <= cloud.numberOfVillagersAttacked then
			self:changeState(JFCloudStateDying)
		else
			self:changeState(JFCloudStateChasing)
		end	
	end
end

--------------------------------------------------------------------------------------------------
--										JFCloudStateDying
--------------------------------------------------------------------------------------------------
waxClass{"JFCloudStateDying", JFCloudState}

function onEnter_withMessage(self, previousState, message)
	local cloud = self:getCloud()
	
	-- set costume
	cloud:wardrobe():playAnimation("FX_CLOUD_IDLE")
	
	-- create pop in action
	local actionScale = CCScaleTo:actionWithDuration_scale(1.0, 0.0)
	local actionFinish = CCCallFunc:actionWithTarget_selector(self, "animationFinished")
	
	-- build sequences
	local sequence = CCSequence:actionOne_two(actionScale, actionFinish)
	
	-- run sequence
	cloud:rootNode():runAction(sequence)
end


function animationFinished(self)
	-- remove cloud from GOS
	local gos = JFServiceLocator:serviceOfType(JFGameObjectService:class())
	gos:removeGameObject(self:getCloud())
end


--]]