include("scripts/enums/JFActorDirection.lua")
include("scripts/enums/JFConsumeTouchPriority.lua")
include("scripts/enums/JMJobType.lua")
include("scripts/enums/JFCommonMessages.lua")
include("scripts/enums/VillagerAgeCategory.lua")

local mapService = JFServiceLocator:serviceOfType(JFMapService:class())
local goService = JFServiceLocator:serviceOfType(JFGameObjectService:class())
local scheduler = CCDirector:sharedDirector():scheduler()


----- the following functions/vars are bug-specific and should be replicated in their own bug file

--------------------------------------------------------------------------------------------------
--                                        Startup function
--------------------------------------------------------------------------------------------------

function JFBugsOnOpen(posx, posy, token)
    local bug = JFBug:initWithModel_fromToken(nil, token)
    bug.payout_number = 1
    bug:spawnAtPosition(CGPoint(posx, posy))
end

--------------------------------------------------------------------------------------------------
--                                        JFBug
--------------------------------------------------------------------------------------------------
waxClass{"JFBug", JFActor, protocols = {"JFWalkingHandler", "JFAudioHandler"}}

nextIdentifier = 1.0

function initWithModel_fromToken_withMinigameInfo(self, model, token, minigameInfo)
    self = self.super:initWithModel(model)
    if self then
        -- assign identifier
        self.identifier = goService:nextUniqueIdentifier()
        
        self.minigame = minigameInfo
        
        self.bugIdentifier=self.minigame.bugIdentifier
        -- generation specific settings
        self.generation = 1
        self.walkSpeed = self.minigame.baseSpeed * (self.minigame.speedScaling^self.generation)
        self:rootNode():setScale(self.minigame.baseSize*self.minigame.sizeScaling^(self.generation-1))
        
        -- audio initialization
        self.persistentId = "minigamebug_" .. self.minigame.bugtype .. self.bugIdentifier
        self.audioHandler = JFAudioComponent:alloc():initWithHandler_andPersistentId(self, self.persistentId)
        
        -- other initialization
        self:setupStateMachine()
        self.walk = JFWalkingComponent:alloc():initWithHandler(self)
        self.scaredVillagers = {}
        
        self:setupHitBox()        
        self:setupTouchHandler()
    end
    return self 
end

function setupHitBox(self)
    -- hitbox initialization
    self.hitbox = JFDebugDrawNode:node()
    self.hitbox:setContentSize(CGSize(35,35))
    self.hitbox:setAnchorPoint(CGPoint(0.5,0.5))
    self.hitbox:setPosition(CGPoint(10,35))
    self:rootNode():addChild(self.hitbox)
end

----- end of bug-specific functions

--[[
-- shows deallocations
testnum = 0 function dealloc(self)  testnum=testnum+1  print("DEALLOCATING #" .. testnum) end
--]]

function initWithModel_fromParent(self, model, parent)
    self = self.super:initWithModel(model)
    if self then
        -- assign identifier
        self.identifier = goService:nextUniqueIdentifier()

        
        self.minigame = parent.minigame
        self.minigame.liveBugs = self.minigame.liveBugs + 1
        self.minigame.bugIdentifier = self.minigame.bugIdentifier+1
        self.bugIdentifier = self.minigame.bugIdentifier

        -- generation specific settings
        self.generation = parent.generation + 1
        self.walkSpeed = self.minigame.baseSpeed * (self.minigame.speedScaling^self.generation)
        self:rootNode():setScale(self.minigame.baseSize*self.minigame.sizeScaling^(self.generation-1))
        
        
        -- audio initialization
        self.persistentId = "minigamebug_" .. self.minigame.bugtype .. self.bugIdentifier
        self.audioHandler = JFAudioComponent:alloc():initWithHandler_andPersistentId(self, self.persistentId)
        
        -- other initialization
        self:setupStateMachine()
        self.walk = JFWalkingComponent:alloc():initWithHandler(self)
        self.scaredVillagers = {}
        
        self:setupHitBox()
        self:setupTouchHandler()
    end
    return self 
end

function getWanderSoundEffect(self)
    if self.generation == 1 then
        return "effectMinigameBugBig"
    elseif self.generation == 2 then
        return "effectMinigameBugMedium"
    else
        return "effectMinigameBugSmall"
    end
end

function setupWanderSound(self)
    self.audioHandler:startPersistentSoundEffect(self:getWanderSoundEffect())
end

function playSplatSound(self)
    local audioInfo = {effectMinigameBugSplat="effect_play", actor=self}
    NSNotificationCenter:defaultCenter():postNotificationName_object_userInfo("call_audio", nil, audioInfo)

    self.audioHandler:stopPersistentSoundEffects()
end

function handleMessage(self, message)
    if self.super.handleMessage == self.handleMessage then
        self.super:handleMessage(message)
        return
    end
    if message:name() == "removeGameObject" then
        self:teardown()
    end
    self.super:handleMessage(message)
end

function classId(self)
    return JFUtility:convertClassToClassId(JFBug:class())
end

function spawnAtPosition(self, tilePosition)
    -- set position
    self:setIsoPos(tilePosition)
    -- add to game object service and map
    goService:addGameObject(self)
    mapService:addActor(self)
    -- put the bug into the invincible wandering state
    self:changeState(JFBugStateInvincible)
    self:setupWanderSound()
end

function changeState(self, stateClass)
    -- convenience function so this doens't have to be written every time
    self:stateMachine():setCurrentState_withMessage(stateClass:class(), nil)
end

function setupStateMachine(self)
    self:stateMachine():addState(JFBugStateScurry:initWithBug(self))
    self:stateMachine():addState(JFBugStateInvincible:initWithBug(self))
    self:stateMachine():addState(JFBugStateSplat:initWithBug(self))
    self:stateMachine():addState(JFBugStateDie:initWithBug(self))
end
    
function walkSpeed(self)
    return self.walkSpeed
end

function updateWalkingCostume(self, aNewDirection)
    self:updateWalkingCostume_force(aNewDirection,force)
end

function updateWalkingCostume_force(self, aNewDirection, force)
    self:faceDirection(aNewDirection)
    local backWalk = self.minigame.bugtype .. "_FLOAT_BACK"
    local frontWalk = self.minigame.bugtype .. "_FLOAT_FRONT"
    if self:shouldFaceFront(aNewDirection) then
        self:wardrobe():playAnimation(frontWalk)
    else
        self:wardrobe():playAnimation(backWalk)
    end
end

function shouldFaceFront(self, aNewDirection)
    if aNewDirection==JFActorDirectionNone or aNewDirection==JFActorDirectionSouth or aNewDirection==JFActorDirectionEast then
        return true
    end
    return false
end

function faceDirection(self, aDirection)
    self.direction = aDirection
    local scaleX = self:rootNode():scaleX()
    if self:shouldBeMirrored() then
        if scaleX > 0 then
            self:rootNode():setScaleX(-scaleX)
        end
    else
        if scaleX < 0 then
            self:rootNode():setScaleX(-scaleX)
        end
    end
end

function shouldBeMirrored(self)
    if self.direction==JFActorDirectionNone or self.direction==JFActorDirectionSouth or self.direction==JFActorDirectionWest then
        return true
    end
    return false
end

function setupTouchHandler(self)
    self.touchHelper = JFTouchHelper:alloc():initWithNode_withOwner(self.hitbox, self)
    self.touchHelper:registerHandler_andSwallowsTouches(CONSUME_TOUCH_PRIORITY_WORLD,true)
    self.touchHelper:setOnTappedCallback(toobjc(function()
                                    self:stateMachine():currentState():handleMessage(JFMessage:messageWithName("touched"))
                                 end):asVoidNiladicBlock())
end

function isVillagerEligible(self, villager)
    if villager:stateMachine() == nil then
        return false
    end
    local villagerState = villager:stateMachine():currentState()
    if villagerState == nil then
        return false
    end
    -- if this actor is in the process of being chased it can't be eligible to be chased a second time
    if villagerState:isKindOfClass(JFVillagerStateAnimate:class()) then
        if villagerState:stateMachine():currentState() == JFVillagerAnimateStateRunaway:class() then
            return false
        end
    end
    local villagerModel = villager:villagerModel()
    -- only adults can be chased
    if villagerModel:ageCategory() ~= VILLAGER_AGE_CATEGORY_ADULT then
        return false
    end
    local villagerJob = villagerModel:job():techtree():type()
    -- if villagers are parenting they cannot be scared
    if villagerJob == VILLAGER_JOB_TYPE_PARENT then
        return false
    end
    -- can't if villagers are hovering or breeding
    if villagerState:isKindOfClass(JFVillagerStateHovering:class()) or villagerState:isKindOfClass(JFVillagerStateBreeding:class()) then
        return false
    end
    -- fair game!
    return true
end

function removeFromGame(self)
    self.minigame.liveBugs = self.minigame.liveBugs - 1
    goService:removeGameObject(self)
end

function teardown(self)
    
    
    
    -- unscare villagers
    if self.scaredVillagers ~= nil then
        for i, villager in ipairs(self.scaredVillagers) do
            villager:stateMachine():setCurrentState_withMessage(JFVillagerStateResumeModelState:class(), nil)
        end
    end
    -- remove minigame from cache if all bugs have died
    if self.minigame.liveBugs == 0 then
        --[[
        local mgs = JFServiceLocator:serviceOfType(JFMiniGameCacheService:class())
        mgs:removeMiniGameByToken(self.minigame.token)
        ]]--
        
        -- unload audio
        local audioInfo = {groupAnimalsBug="unload_group"}
        NSNotificationCenter:defaultCenter():postNotificationName_object_userInfo("call_audio", nil, audioInfo)
    end
    
    self:rootNode():removeFromParentAndCleanup(true)

    self:stateMachine():destroy()
    self.minigame = nil
    
    self.touchHelper:cleanup()
    self.touchHelper = nil
    self.walk:cleanup()
    self.audioHandler:cleanup()
    self:wardrobe():clear()
end


--------------------------------------------------------------------------------------------------
--                                        JFBugStateBase
--------------------------------------------------------------------------------------------------
waxClass{"JFBugStateBase", JFStateBase}

function initWithBug(self, bugz)
    self = self.super:init()
    if self then
        self.bug = bugz
    end
    return self
end

function getBug(self)
    return self.bug
end


--------------------------------------------------------------------------------------------------
--                                        JFBugStateScurry
--------------------------------------------------------------------------------------------------
waxClass{"JFBugStateScurry", JFBugStateBase}

function onEnter_withMessage(self, previousState, message)
    self.super:onEnter_withMessage(previousState, message)
    self:scurryAround()
    -- setup scheduler tasks
    scheduler:scheduleSelector_forTarget_interval_paused_repeat_delay("scareVillagers", self,  self.bug.minigame.villagerScareInterval, false, kCCRepeatForever, 0)
    local expireInterval = self.bug.minigame.expiryDate:timeIntervalSinceNow()
    if expireInterval < 0 then
        expireInterval = 0
    end
    scheduler:scheduleSelector_forTarget_interval_paused_repeat_delay("expire", self, 0, false, 0, expireInterval)
end

function scurryAround(self)
    local radius = self.bug.minigame.wanderRadius
    
    self.bug.walk:startRandomWalkWithMaxRadius_minRadius(radius, radius/4.0)
end

function handleMessage(self, aMessage)
    if aMessage:name() == "touched" then
        self.super:handler():setCurrentState_withMessage(JFBugStateSplat:class(), aMessage)
    end
    self.super:handleMessage(aMessage)
end

function scareVillagers(self)
    local villagers = mapService:getGameObjectsInRadius_fromPoint_withClassType(self.bug.minigame.villagerScareRadius, self.bug:isoPos(), JFVillager:class())
    for i, villager in ipairs(villagers) do
        if self.bug:isVillagerEligible(villager) then
            table.insert(self.bug.scaredVillagers, villager)
            local message = JFMessage:messageWithName("runningAway")
            local messageData = {substate=JFVillagerAnimateStateRunaway:class()}
            message:setData(messageData)
            villager:stateMachine():setCurrentState_withMessage(JFVillagerStateAnimate:class(), message)
        end
    end 
end

function expire(self)
    self.bug.generation = self.bug.minigame.maxNumberGenerations -- make sure no future generations spawn
    self.bug:changeState(JFBugStateSplat)
end

function onExit_withMessage(self, previousState, message)
    self.super:onExit_withMessage(previousState, message)
    self.bug.walk:stopWalking()
    
    -- remove scheduler tasks
    scheduler:unscheduleAllSelectorsForTarget(self)
end


--------------------------------------------------------------------------------------------------
--                                        JFBugStateInvincible
--------------------------------------------------------------------------------------------------
waxClass{"JFBugStateInvincible", JFBugStateScurry}

function onEnter_withMessage(self, previousState, message)
    self.super:onEnter_withMessage(previousState, message)
    scheduler:scheduleSelector_forTarget_interval_paused_repeat_delay("becomeMortal", self, 0, false, 0, self.bug.minigame.invincibleDuration)
end

function handleMessage(self, aMessage)
    if aMessage:name() == "touched" then
        return
    else
        self.super:handleMessage(aMessage)
    end
end

function becomeMortal(self)
    self.bug:changeState(JFBugStateScurry)
end

--------------------------------------------------------------------------------------------------
--                                        JFBugStateSplat
--------------------------------------------------------------------------------------------------
waxClass{"JFBugStateSplat", JFBugStateBase}

function onEnter_withMessage(self, previousState, message)
    self.super:onEnter_withMessage(previousState, message)
    
    if (self.bug.generation < self.bug.minigame.maxNumberGenerations) then
        local nextGeneration = 2^self.bug.generation
        for i=1, nextGeneration do
            local newBug = self.bug:class():initWithModel_fromParent(nil, self.bug)
            newBug.payout_number=(self.bug.payout_number-1)*nextGeneration + i
            newBug:spawnAtPosition(self.bug:isoPos())
        end
    end
    
    self.bug:playSplatSound()    
    self.splatAnimation = KAnimation:animationWithName_target_collections("FX_BUG_SPLAT", self, nil)
    self.splatAnimation:addCallbackOnFinished_withSelector(self, "onFinished:withAnimation:")

    self.bug:rootNode():addChild(self.splatAnimation)
    self.splatAnimation:play_forward_paused(false, true, false)
    self.bug:wardrobe():clear()
    self.bug:updateSortingDepth()
end

function onFinished_withAnimation(self, animationName, anAnimation)
    if anAnimation == self.splatAnimation then
        self.bug:changeState(JFBugStateDie)
    end
end

function onExit_withMessage(self, previousState, message)
    self.super:onExit_withMessage(previousState, message)
    self.splatAnimation:removeAllCallbacks()
end

--------------------------------------------------------------------------------------------------
--                                        JFBugStateDie
--------------------------------------------------------------------------------------------------
waxClass{"JFBugStateDie", JFBugStateBase}

function onEnter_withMessage(self, previousState, message)
    self.super:onEnter_withMessage(previousState, message)
    self.bug:wardrobe():clear()
    self.bug:removeFromGame()
end


function onExit_withMessage(self, previousState, message)
    self.super:onExit_withMessage(previousState, message)
end