include("scripts/enums/JFActorDirection.lua")
include("scripts/enums/JMJobType.lua")
include("scripts/enums/VillagerAgeCategory.lua")
include("scripts/soccer/VillagerStatePlayingSoccer.lua")

--------------------------------------------------------------------------------------------------
--										Startup function
--------------------------------------------------------------------------------------------------

function JFSoccerBallOnOpen(posx, posy)
	local soccerball = JFSoccerBall:initWithModel(nil)
	soccerball:spawnAtPosition(CGPoint(posx, posy))
end

function JFSoccerBallOnRemove()
	local gos = JFServiceLocator:serviceOfType(JFGameObjectService:class()) -- game object service
	local soccerballs = gos:gameObjectsWithClassType(JFSoccerBall:class())
    
    if soccerballs ~= nil then
		for i, soccerball in ipairs(soccerballs) do
			soccerball:removeBallFromPlay()
		end
	end
end

function JFSoccerBallOnReOpen(posx, posy)
    JFSoccerBallOnRemove()
    JFSoccerBallOnOpen(posx, posy)
end

--------------------------------------------------------------------------------------------------
--										JFSoccerBall
--------------------------------------------------------------------------------------------------
nextSoccerBallIdentifier = 1.0

waxClass{"JFSoccerBall", JFActor}

-- time before ball disappears (5 mins)
lifetime = 300

-- interval to check if the children are stuck
forceKickInterval = 10

function initWithModel(self, model)
	self = self.super:initWithModel(model)
	if self then
		--self.owner = owner
		self.identifier = nextSoccerBallIdentifier;
		nextSoccerBallIdentifier = nextSoccerBallIdentifier + 1.0
		
		-- initially allow the ball to be kicked
		self.canBeKicked = true;
		self.playDistance = 132.0
		
		self:setupWardrobe()
		
		self.claimedVillager = -1
		
		self.childrenPlayingWith = 0
        
        -- when the ball was kicked last time, used for hacking
        -- init to now() + 20s
        self.dateLastKicked = NSDate:dateWithTimeIntervalSinceNow(20)
	end
	return self
end


function setupWardrobe(self)
	-- set soccer ball animation
	self:wardrobe():playAnimation("WI_SOCCER_BALL_STATIC")
	
	-- need to move the offsetnode position up slightly
	self:offsetNode():setPosition(CGPoint(0.0, 6.0))
	
	-- add shadow to the soccer ball
	local shadowSpriteName = "SHADOW_STOCK.png"
	local shadow = CCSprite:spriteWithSpriteFrameName(shadowSpriteName)
	shadow:setScale(0.65) -- scale shadow down, otherwise it's huge
	self:rootNode():addChild_z(shadow, -1)
end


function spawnAtPosition(self, tilePosition)
	-- get the services requried
	local gos = JFServiceLocator:serviceOfType(JFGameObjectService:class()) -- game object service
	local map = JFServiceLocator:serviceOfType(JFMapService:class()) -- map service
	
	self:setIsoPos(tilePosition)
	
	-- add soccer ball to game object service
	gos:addGameObject(self)
	
	-- add the soccer ball to the map
	map:addActor(self)
	
	print("Spawned soccer ball at " .. tostring(tilePosition))
	
	self.nextPosition = tilePosition
	self.prevPosition = tilePosition
	self:calculateNextPosition()
	
	self:lureChildren()
    
    self:performSelector_withObject_afterDelay("testAndForceLureChildren", nil, forceKickInterval)

	-- game over after 5 minutes :(
    -- will removed by VillageService
	-- self:performSelector_withObject_afterDelay("removeBallFromPlay", nil, lifetime)
end


function classId(self)
    return JFUtility:convertClassToClassId(JFSoccerBall:class())
end


function ident(self)
	return self.identifier
end


function handleMessage(self, message)
	self.super:handleMessage(message)
	
	if message:name() == "removeGameObject" then 
		print("Removing soccer ball game object")
        self:stopChildrenFromPlaying()
		NSObject:cancelPreviousPerformRequestsWithTarget(self)
	end
end


function calculateNextPosition(self)
	self.prevPosition = self.nextPosition
	
	-- choose a random direction
	local map = JFServiceLocator:serviceOfType(JFMapService:class(), nil)
	
	local randomPosition = map:getRandomSpawnPoint()
	
	-- get direction to kick from the random spanw point
	local difference = CGPoint(randomPosition.x - self.prevPosition.x, randomPosition.y - self.prevPosition.y)
	local dir = JFActor:getDirectionFromPoint(difference)
	
	-- get the normalized tile position for kicking
	local offset = JFActor:getOffsetFromDirection(dir)
	
	-- add the position to the villager position to get the absolute position relative to villager
	local kickDistanceMax = 10 -- maximum number of tiles away the kick should go
	local kickDistanceMin = 4  -- minimum number
	
	-- get the map service
	local map = JFServiceLocator:serviceOfType(JFMapService:class(), nil)
	
	-- pop, POP
	local magnitude = map:randomFloatBetween_between(kickDistanceMin, kickDistanceMax)
	
	offset.x = self.prevPosition.x + (offset.x * magnitude)
	offset.y = self.prevPosition.y + (offset.y * magnitude)
	
	-- make sure the position selected by the ball is a valid position to run to
 	self.nextPosition = map:getPointByWindingOutFromTile(offset)
end


function getCanBeKicked(self)
	return self.canBeKicked
end


function getNextPosition(self)
	return self.nextPosition
end


function getPrevPosition(self)
	return self.prevPosition
end


function getPlayDistance(self)
	return self.playDistance
end


function onLanding(self)
	-- make sure that the ball can not be kicked when it lands
	if self.canBeKicked == false then
		-- ball can be kicked again
		self.canBeKicked = true
		self.claimedVillager = -1 -- no longer claimed
		
		self:lureChildren()
	end
end


function kick(self, fromPosition, toPosition)
	-- make sure the ball can be kicked
	if self.canBeKicked == true then
        
        -- get kicked now
        self.dateLastKicked = NSDate:date()

		self:calculateNextPosition() -- figure out where it's going to go next
		
		-- wait time for the "kick"
		local waitTime = 0.2
		local wait = CCDelayTime:actionWithDuration(waitTime)
		
		-- configure how fast the ball rolls
		local secondsPerTile = 0.1
		
		-- get the screen position of the ball
		local map = JFServiceLocator:serviceOfType(JFMapService:class(), nil)
		local ballScreenPosition = map:tilePositionToScreenPosition(toPosition)
		
		local difference = CGPoint(fromPosition.x - toPosition.x, fromPosition.y - toPosition.y)
		local magnitude = math.sqrt(difference.x * difference.x + difference.y * difference.y)
		local dir = JFActor:getDirectionFromPoint(difference)
		self.lastMagnitude = magnitude
		
		local duration = secondsPerTile * magnitude
		local halfDuration = duration / 2.0

		-- bounce the ball
		do
			local bounceHeight = 8.0 * magnitude -- how high the ball should bounce

			-- offset node height of kick and original positions
			local offsetNodePositionOriginal = self:offsetNode():position()
			local offsetNodePositionBouncePeak = CGPoint(offsetNodePositionOriginal.x, offsetNodePositionOriginal.y + bounceHeight)
			
			-- ease out to look floaty
			local upAction = CCEaseExponentialOut:actionWithAction(CCMoveTo:actionWithDuration_position(halfDuration, offsetNodePositionBouncePeak))
			-- bounce to look bouncy 
			local downAction = CCEaseBounceOut:actionWithAction(CCMoveTo:actionWithDuration_position(halfDuration, offsetNodePositionOriginal))
			-- sequence of wait, then up and down
			local sequence = CCSequence:actionOne_two(wait, CCSequence:actionOne_two(upAction, downAction))

			self:offsetNode():runAction(sequence)
		end

		-- rotate the ball
		do
			local spinDirection = 1.0
			local spinSpeed = 1080.0 -- degrees per second

			if dir == JFActorDirectionSouth or dir == JFActorDirectionWest then
				spinDirection = -1.0
			end

			local spinAction = CCSequence:actionOne_two(wait, CCRotateBy:actionWithDuration_angle(duration, duration * spinSpeed * spinDirection))
			self:offsetNode():runAction(spinAction)
		end
		
		-- move the ball to the desired screen position
		do
			local roll = CCMoveTo:actionWithDuration_position(duration, ballScreenPosition)
			local callfunc = CCCallFunc:actionWithTarget_selector(self, 'onLanding')
			local rollcall = CCSequence:actionOne_two(roll, callfunc)
			local waitroll = CCSequence:actionOne_two(wait, rollcall)
			
			self:rootNode():runAction(waitroll)
		end

		-- play bounced sounds
		do
			local firstWait = CCDelayTime:actionWithDuration(waitTime + duration*.60)
			local firstCallFunc = CCCallFunc:actionWithTarget_selector(self, 'playFirstBounceSound')
			local secondWait = CCDelayTime:actionWithDuration(duration*.1)
			local secondCallFunc = CCCallFunc:actionWithTarget_selector(self, 'playSecondBounceSound')
			local firstBounce = CCSequence:actionOne_two(firstWait, firstCallFunc)
			local secondBounce = CCSequence:actionOne_two(secondWait, secondCallFunc)
			local bounceSequence = CCSequence:actionOne_two(firstBounce, secondBounce)
			self:rootNode():runAction(bounceSequence)
		end

		-- can only be kicked one time
		self.canBeKicked = false 
		
		-- send a message that the ball was kicked so villagers who didn't kick the ball can handle it!
		-- play soccer with this ball instead of a new one
		messageData = { }
		messageData["soccerball_id"] = self.identifier
		message = JFMessage:messageWithName("soccerBallKicked")
		message:setData(messageData)
		message:dispatch()

	end
end

function playFirstBounceSound(self)
	notificationData = { }
	notificationData["effectKickballBounce"] = "effect_play"
	notificationData["actor"] = self
	notificationData["parameter_value"] = self.lastMagnitude

	NSNotificationCenter:defaultCenter():postNotificationName_object_userInfo("call_audio", nil, notificationData)
end

function playSecondBounceSound(self)
	notificationData = { }
	notificationData["effectKickballBounce"] = "effect_play"
	notificationData["actor"] = self
	notificationData["parameter_value"] = self.lastMagnitude*0.5

	NSNotificationCenter:defaultCenter():postNotificationName_object_userInfo("call_audio", nil, notificationData)
end

function lureChildren(self)
	gos = JFServiceLocator:serviceOfType(JFGameObjectService:class()) -- game object service
	local villagers = gos:gameObjectsWithClassType(JFVillager:class())
    self.canBeKicked = true
	local numPlayers = 0
    
	if villagers ~= nil then
		for i, villager in ipairs(villagers) do
			local isChild = villager:villagerModel():ageCategory() == VILLAGER_AGE_CATEGORY_CHILD and (villager:villagerModel():job():techtree():type() == VILLAGER_JOB_TYPE_CHILD or villager:villagerModel():job():techtree():type() == VILLAGER_JOB_TYPE_IDLE)
			if isChild and villager ~= self.owner then
				local childPosition = villager:isoPos()
				local position = self.prevPosition
		            
                local state = villager:stateMachine():currentState()
            
                local isHovering = state:isKindOfClass(JFVillagerStateHovering:class()) == true

                if(villager:stateMachine():stateOfType(VillagerStatePlayingSoccer:class()) == nil) then
                    local state = VillagerStatePlayingSoccer:initWithVillager(villager)
                    villager:stateMachine():addState_forClass(state, VillagerStatePlayingSoccer:class())
                    print("ADDING STATE")
                end

                local isPlayingSoccer = state:isKindOfClass(VillagerStatePlayingSoccer:class()) == true
            
                -- don't lure children that are being dragged
                if isHovering == false then
                    -- if they're not already playing soccer then we need to make them
                    if isPlayingSoccer == false then
                        -- play soccer with this ball instead of a new one
                        messageData = { }
                        messageData["soccerball"] = self
                        message = JFMessage:messageWithName("chaseSoccerBall")
                        message:setData(messageData)
                        villager:stateMachine():setCurrentState_withMessage(VillagerStatePlayingSoccer:class(), message)
                    end
                
                            if state.soccerball == self then
                        numPlayers = numPlayers + 1
                    end
                end
            end
		end
	end

	if numPlayers == 0 then
		self:performSelector_withObject_afterDelay("lureChildren", nil, 3)
	end
end

function testAndForceLureChildren(self)
    local gos = JFServiceLocator:serviceOfType(JFGameObjectService:class()) -- game object service
    local soccerballs = gos:gameObjectsWithClassType(JFSoccerBall:class())
    
    io.write("Checking if the children are stuck\n")
    
    if soccerballs ~= nil and table.getn(soccerballs) > 0 then
        -- there is a kick ball in the village
        if NSDate:date():timeIntervalSinceDate(self.dateLastKicked) > forceKickInterval then
            -- haven't been kicked for a long time, maybe children are stuck
            -- send onWalkComplete message to every children with VillagerStatePlayingSoccer state
            io.write("Children are stuck\n")
            local villagers = gos:gameObjectsWithClassType(JFVillager:class())
            if villagers ~= nil then
                for i, villager in ipairs(villagers) do
                    local isChild = villager:villagerModel():ageCategory() == VILLAGER_AGE_CATEGORY_CHILD and (villager:villagerModel():job():techtree():type() == VILLAGER_JOB_TYPE_CHILD or villager:villagerModel():job():techtree():type() == VILLAGER_JOB_TYPE_IDLE)
                    
                    if isChild and villager:stateMachine() ~= nil and villager:stateMachine():currentState() ~= nil and villager:stateMachine():currentState():isKindOfClass(VillagerStatePlayingSoccer:class()) then
                        io.write("Sending walkComplete message to a child\n")
                        local message = JFMessage:messageWithName("walkComplete")
                        villager:handleMessage(message)
                    end
                end
            end
        end
        
        self:performSelector_withObject_afterDelay("testAndForceLureChildren", nil, forceKickInterval)
    end
end

function removeBallFromPlay(self)
	print("Removing soccer ball :(")

    NSObject:cancelPreviousPerformRequestsWithTarget(self)

	self:stopChildrenFromPlaying()

	-- get the services requried
	local gos = JFServiceLocator:serviceOfType(JFGameObjectService:class()) -- game object service
    
	-- remove the soccer ball to the map
	--self:rootNode():removeFromParentAndCleanup(true)

	-- remove soccer ball to game object service
	gos:removeGameObject(self)
	
end

function stopChildrenFromPlaying(self)
	local gos = JFServiceLocator:serviceOfType(JFGameObjectService:class()) -- game object service
	local villagers = gos:gameObjectsWithClassType(JFVillager:class())
    
	local numPlayers = 0

	if villagers ~= nil then
		for i, villager in ipairs(villagers) do
			if(villager:stateMachine():stateOfType(VillagerStatePlayingSoccer:class()) ~= nil) then
				local state = villager:stateMachine():currentState()
				if state:isKindOfClass(VillagerStatePlayingSoccer:class()) == true and state.soccerball == self then
					villager:stateMachine():setCurrentState_withMessage(JFVillagerStateIdle:class(), nil)
					print(villager:villagerModel():name() .. " stopped playing soccer")
				end
			end
		end
	end
end



--------------------------------------------------------------------------------------------------
--										Spawn a soccer ball
--------------------------------------------------------------------------------------------------


-- takes care of adding the cloud as sender
function createMessage(self, messageName, messageData)
	
	assert(messageName ~= nil)
	assert(self.cloud ~= nil)
	
	if messageData == nil then
		messageData = { }
	end
	
	messageData["sender"] = self.cloud
	
	local message = JFMessage:messageWithName(messageName)
	message:setData(messageData)
	
	return message
	
end


function sendMessageToVillager(self, messageName, messageData)

	local message = self:createMessage(messageName, messageData)
	
	local villager = self:getCloud():getVillager()
	villager:handleMessage(message)
	
end