include("scripts/enums/JFActorDirection.lua")

--------------------------------------------------------------------------------------------------
--							VillagerStatePlayingSoccer
--------------------------------------------------------------------------------------------------

waxClass{"VillagerStatePlayingSoccer", JFVillagerDynamicState}

function initialize(self)	
	-- create the substates
	local stateDropBall   = VillagerPlayingSoccerDropBall:initWithParentState(self)
	local stateKickBall   = VillagerPlayingSoccerKickBall:initWithParentState(self)
	local stateChaseBall  = VillagerPlayingSoccerChaseBall:initWithParentState(self)
	
	-- add substates to the state machine
	self:stateMachine():addState(stateDropBall)
	self:stateMachine():addState(stateKickBall)
	self:stateMachine():addState(stateChaseBall)
end


function onEnter_withMessage(self, previousState, message)
	self.super:onEnter_withMessage(previousState, message)

	self.timer = 0.0

	self:applyCostume()
	
	if message ~= nil and message:name() == "chaseSoccerBall" then
		-- if the message is not nil and says to chase the ball, then use the existing ball passed in
		local messagedata = message:data()
		self.soccerball = messagedata["soccerball"]
		self:stateMachine():setCurrentState_withMessage(VillagerPlayingSoccerChaseBall:class(), nil)
	else
		-- spawn soccer ball if no ball was specified to play with
		self.soccerball = JFSoccerBall:initWithModel_andOwner(nil, self:villager())
		self.soccerball:spawnAtPosition(self:villager():isoPos())
		-- go into the dropping ball substate
		self:stateMachine():setCurrentState_withMessage(VillagerPlayingSoccerDropBall:class(), nil)
	end
	
	self.soccerball.childrenPlayingWith = self.soccerball.childrenPlayingWith + 1
    
    self:villager():loadCostumeFromFactory("CHILD_F_KICK")
    self:villager():loadCostumeFromFactory("CHILD_B_KICK")

    self:villager():walk():setForceRun(true)
end

function onExit_withMessage(self, nextState, message)
	self.super:onExit_withMessage(nextState, message)
	
	self.soccerball.childrenPlayingWith = self.soccerball.childrenPlayingWith - 1
	-- clean up the soccer ball from the game
	if self.soccerball.childrenPlayingWith <= 0 then
		self.soccerball:performSelector_withObject_afterDelay("lureChildren", nil, 5)
	end

    self:villager():walk():setForceRun(false)
end

function canExitToState_withMessage(self, nextState, message)
	if nextState:isKindOfClass(JFVillagerStateStudent:class()) then
		return true
	end
	return self.super:canExitToState_withMessage(nextState, message)
end


------------------------------------------------------------------------------------------------
--							VillagerPlayingSoccerDropBall
------------------------------------------------------------------------------------------------

-- we might want this state in the future, but right now it only calculates the first position to kick to
-- and then kicks the ball
waxClass{"VillagerPlayingSoccerDropBall", JFVillagerDynamicState}

function onFirstRun_withMessage(self, previousState, message)
	self:parentState():stateMachine():setCurrentState_withMessage(VillagerPlayingSoccerKickBall:class(), nil)
end

--------------------------------------------------------------------------------------------------
--							VillagerPlayingSoccerKickBall
--------------------------------------------------------------------------------------------------

waxClass{"VillagerPlayingSoccerKickBall", JFVillagerDynamicState}

function applyCostume(self)
	-- apply the default costume so that the child doesn't continue to run when they reach
	-- the ball in the split second before kicking it
	local v = self:parentState():villager()
	v:wardrobe():playAnimation("CHILD_IDLE_START")
end


function onEnter_withMessage(self, previousState, message)
	self.super:onEnter_withMessage(previousState, message)

	-- set timer to zero and kicked ball to false
	self.timer = 0.0
	self.kickedBall = false

	self:applyCostume()
	
	-- kick the ball immediately
	if self:parentState().soccerball:getCanBeKicked() == true then
		self:kickBall()
	else
		-- can't kick the ball, so just go back to chasing
		self.timer = 99.0
		self.kickedBall = true
		io.write("IN KICK STATE, BUT CAN'T KICK!!\n")
	end
end


function update(self, elapsedSeconds)
	-- update timer and base state
	self.super:update(elapsedSeconds)
	
	-- update the sorting depth of the ball since we are in charge of and it will be moving
	self:parentState().soccerball:updateSortingDepth() -- shouldn't be here
	
	-- if the ball was kicked then wait an additional delay before chasing
	if self.kickedBall == true then
		local chaseDelay = 0.45	-- seconds after kicking the ball to chase it
		self.timer = self.timer + elapsedSeconds
		if self.timer > chaseDelay then
			self:parentState():stateMachine():setCurrentState_withMessage(VillagerPlayingSoccerChaseBall:class(), nil)
		end
	end
end


function kickBall(self)
	-- get the tile and screen position for the ball's landing
	local ballTilePosition = self:parentState().soccerball.nextPosition
	
	-- get villager from parent and position
	local v = self:parentState():villager()
	local villagerPosition = v:isoPos()
	
	-- get the direction of the final tile
	local difference = CGPoint(ballTilePosition.x - villagerPosition.x, ballTilePosition.y - villagerPosition.y)
	local dir = JFActor:getDirectionFromPoint(difference)
		
	-- play the kick animation
	local option = {}
	option["Repeat"] = "NO";
	local animationName = "CHILD_F_KICK"
	if dir == JFActorDirectionNorth or dir == JFActorDirectionWest then
		animationName = "CHILD_B_KICK"
	end

	-- face direction and set costume
	v:faceDirection(dir)
	v:wardrobe():playAnimation_withOption(animationName, option)
	
	self:parentState().soccerball:kick(villagerPosition, ballTilePosition)
		
	-- did kick the ball
	self.kickedBall = true
	self.timer = 0.0
end


--------------------------------------------------------------------------------------------------
--							VillagerPlayingSoccerChaseBall
--------------------------------------------------------------------------------------------------

waxClass{"VillagerPlayingSoccerChaseBall", JFVillagerDynamicState}

function onEnter_withMessage(self, previousState, message)
	self.super:onEnter_withMessage(previousState, message)
end


function onFirstRun_withMessage(self, previousState, message)
	self.super:onFirstRun_withMessage(previousState, message)
	self:runToBall()
end


function runToBall(self)	
	local parent = self:parentState()
	local villager = parent:villager()
	
	local currentBallPosition = self:parentState().soccerball:getPrevPosition()
	local nextBallPosition = self:parentState().soccerball:getNextPosition()
	
	-- this is intentionally backwards to give the opposite direction of the kick (to walk to the other side of ball)
	local difference = CGPoint(currentBallPosition.x - nextBallPosition.x, currentBallPosition.y - nextBallPosition.y)
	local dir = JFActor:getDirectionFromPoint(difference)
	
	local offset = JFActor:getOffsetFromDirection(dir)
	
	local villagerPosition = CGPoint(currentBallPosition.x + (offset.x * 1.0), currentBallPosition.y + (offset.y * 1.0))
	
	-- if the tile the villager should go to is occupied, just go to the ball
	-- this can happen if the ball lands next to an occupied tile
	local map = JFServiceLocator:serviceOfType(JFMapService:class(), nil)
	if map:isTileOccupied(villagerPosition) then
		villagerPosition = currentBallPosition
	end
	
	local visopos = villager:isoPos()
	local vdiff = CGPoint(visopos.x - villagerPosition.x, visopos.y - villagerPosition.y)
	vdiff = math.sqrt(vdiff.x*vdiff.x, vdiff.y*vdiff.y)
	
	local playDist = self:parentState().soccerball:getPlayDistance()
	
	if vdiff <= playDist then
		-- if the villager is still in play distance then chase the ball
		-- io.write(villager:villagerModel():name() .. " villager walking from: ", visopos.x, ", ", visopos.y, "\n")
		-- io.write("to: ", villagerPosition.x, ",", villagerPosition.y, "\n")			
		villager:walk():walkToTile_onArrive(villagerPosition, nil)
	else
		-- if the villager is too far away they should go idle
		--villager:stateMachine():setCurrentState_withMessage(JFVillagerStateIdle:class(), nil)
	end
end


function handleMessage(self, message)
	if message:name() == "walkComplete" then
        self:parentState():stateMachine():setCurrentState_withMessage(VillagerPlayingSoccerKickBall:class(), nil)
	elseif message:name() == "soccerBallKicked" then
		local v = self:parentState():villager()
		
		local messagedata = message:data()
		local ballid = messagedata["soccerball_id"]
        
		if ballid == self:parentState().soccerball:ident() then
			if v:walk():isWalking() then
				v:walk():stopWalking()
			end
			self:runToBall()
		end
	else
		self.super:handleMessage(message)
	end
end


function update(self, elapsedSeconds)
	self.super:update(elapsedSeconds)
	
	-- update the sorting depth of the ball since we are in charge of it
	self:parentState().soccerball:updateSortingDepth() -- shouldn't be here
end
